/*
reziproke Frequenzmessung mit RP2040

Das Eingangssignal an FIN1_PIN wird mit SystemCoreClock/4 abgetastet und
die Zyklen mit dem X-Register gezaehlt. Bei einer fallenden Flanke wird der 
Inhalt des PIO0-SM0-X-Registers auf RXF0 ausgegeben und per DMA in die Variable 
F1_time_stamp geschrieben. Die Anzahl der Zeitstempel wird mit dem Register
DMA_TRANS_COUNT gezaehlt.

Nach Ablauf des vorgegebenen Messintervalls (DEF_F1_MESSZEIT) werden der
Zeitpunkt (F1_time_stamp) und die Anzahl der Ereignisse (DMA_TRANS_COUNT)
ausgelesen und die Differenz zu den zuvor ausgelesen Werten gebildet. Die Zeit 
ist dabei synchron zu den Ereignissen. Die Frequenz ergibt sich als:
Ereignisse/Zeit * Abtastrate.

Die Abtastrate betraegt mit SystemCoreClock/4 typisch 31,25 MHz. Bei dem
vorgegebenen Messintervall von 3 Messungen/s ist jede Messungen auf <= 1*10^-7 
(3 / 3,75*10^7) aufgeloest, was eine 7-stellige Aufloesung der Messwerte ergibt.
Die absolute Genauigkeit haengt von der Genauigkeit und Stabilitaet von 
SystemCoreClock ab.
Zur Korrektur dieser Frequenz dient die Variable 'f_ref_korrektur', die mit
einem Trimmpoti am ADC0-Eingang eingestellt werden kann.

Die uebrigen Funktionen sollten aus den Kommentaren erkennbar sein.

DMA-Trigger geaendert: CH0 und CH10 triggern sich gegenseitig
2023-03-04

update auf Arduino IDE:
2024-02-08

Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include "rp2040x.h"
#include <string.h>
#include <stdio.h>
#include "pico_fmeter.h"

#define DMA_COUNT         0x80000000
#define DMA_COUNT_MASKE   (DMA_COUNT-1)

#define DEF_STELLEN       7
#define DEF_FAKTOR        1
#define DEF_UPM_DIVISOR   1

#define TEMP_STR_LEN      50          // fuer Ergebnis String
#define MAX_ERGEBNIS_STR  20          // Ausgabelaenge begrenzen

extern double f_ref_korrektur;        // +/- 20 ppm vom ADC
extern uint32_t SystemCoreClock;      // je nach Einstellung in der IDE
extern uint32_t *irq_tab;             // IRQ-Vektoren
extern uint8_t lcd_breite;            // typ. 16 optional 20

extern void  warten(uint32_t ms);

extern void  lcd_zeile1(const char *s);
extern void  lcd_zeile2(const char *s);
extern void  loesche_lcd_zeile12();
extern void  lcd_messwert_zeile1(char *s);
extern void  lcd_messwert_zeile2(char *s);

double rundung[] = {5e0, 5e-1,5e-2,5e-3,5e-4,5e-5,
                    5e-6,5e-7,5e-8,5e-9,5e-10,5e-11,
                    5e-12,5e-13,5e-14,5e-15};

char f_dim[][10][6]={{" GHz"," MHz"," kHz"," Hz "," mHz"," sek"," ms "," us "," ns "," ps "},
                    {"E+9","E+6","E+3","E+0","E-3","E+0","E-3","E-6","E-9","E-12"}};  // Dimensionen zum Messwert
char u_dim[][5][6]={{" Grpm"," Mrpm"," krpm"," rpm"," mrpm"},
                    {"E+9","E+6","E+3","E+0","E-3"}};

enum anzeige {WANDEL_F=0, WANDEL_P, WANDEL_D};
enum zeige {F_UND_P=0, FREQUENZ, PERIODE, DREHZAHL};
enum ausgabe_kanal {F1=0, P1, UPM1};  // fuer ser. Datenausgabe

uint32_t  F1_time_stamp,              // Zeitpunkt von PIO0-SM0
          dma_dummy;

uint32_t  led_count,
          led_zeit = TYP_LED_ZEIT;    // Blinkzeit

int8_t    ausgabe_sperre,             // keine LCD Datenausgage
          anzeige_funktion,           // F+P, nur F, nur P, nur UPM oder auch aus -1
          update_LCD,                 // LCD neu beschriften
          nur_MHz,                    // festes Format x.xxx MHz
          anzeige_format,             // bit0: Dimension oder Exponent; bit1: . oder , 
          F1_vorteiler_aktiv,         // nicht verwendet
          ser_ausgabe_kanal;          // ser. Datenausgabe: F, P, UPM oder auch aus -1

char p_str[TEMP_STR_LEN];

double              F1_ergebnis;                    // in Hz

int32_t             F1_ref_offset = 0;              // fuer Abgleich von F1_ref
uint32_t            F1_messzeit = DEF_F1_MESSZEIT,  // Sollintervall
                    F1_timeout = DEF_TIMEOUT,       // Abbruch bei fehlendem Signal
                    F1_stellen = DEF_STELLEN,       // hier konstant 7
                    F1_faktor  = DEF_FAKTOR,        // hier konstant 1
                    F1_upm_divisor  = DEF_UPM_DIVISOR;  // ebenso
                    
volatile uint8_t    F1_messwert_vorhanden;          // gueltige Messung
                                   
volatile uint32_t   F1_mess_dauer,                  // in ms
                    F1_timeout_counter;             // in ms
uint32_t            F1_start_zeit,                  // rel. Zeit: Beginn der Messung
                    F1_end_zeit,                    // Zeitpunkt der Auswertung
                    F1_mess_zeit;                   // genaue Zeit der Messung
                    
uint32_t            F1_mess_ereignisse,             // Impulse der Messung
                    F1_start_ereignis,              // Impulse zu Beginn der Messung
                    F1_end_ereignis;                // Impulse am Ende der Messung



// bedingte Anzeige von Messwerten
void messwert_lcd1(char *s)
{
  if(!ausgabe_sperre) lcd_messwert_zeile1(s);
}  

void messwert_lcd2(char *s)
{
  if(!ausgabe_sperre) lcd_messwert_zeile2(s);
} 

// Routine zur Wandlung und Ausgabe eines Messwertes
void wandel_Fx(const char *header, double x, char zeige_periode, char *s, uint8_t stellen)    // Anzeige von F, P oder UPM
{
int i,j,dez_punkt,dimension;
char *p;
  if(zeige_periode == WANDEL_F && nur_MHz)
  {
    if(x >= 1e6) stellen--;
    if(x >= 1e7) stellen--;
    if(x >= 1e8) stellen--;
    if(x >= 1e9) stellen--;
    sprintf(s,"%.*f MHz",stellen,x*1e-6);   // festes Format bei "nur MHz"
    if(*s == '0') *s = ' ';
    return;
  }    

  dez_punkt = 0;
  dimension = 3;
  p = (char *)header;
  while(*p) *s++ = *p++;                    // "P = " fuer Periodendauer
  if(x >= 0.001) {                          // 1mHz ist Untergrenze
    if(zeige_periode == WANDEL_P) {
      x = 1/x;                              // Kehrwert bilden
      dimension=5;                          // und in Sekunden als Dimension
    }
    while(x<1.0) {x*=1000.0;dimension++;}   // in den Hz-Bereich bringen
    while(x>=1000.0) {x*=0.001;dimension--;}
    while(x >= 10.0) {
      x *= 0.1;
      dez_punkt++;
    }
    x += rundung[stellen];                  // runden
    if(x >= 10.0) {			                    // Ueberlauf bei Rundung
      x *= 0.1;                             // korrigieren
      dez_punkt++;
      if(dez_punkt > 2) {		                // Ueberlauf der Dimension
        dimension--;
        dez_punkt = 0;
      }
    }
  } else {
    x = 0.0; dez_punkt=0; dimension=3;      // 0.00000 Hz ausgeben
  }
  
  for(i=0;i<stellen;i++) {
    j = (uint8_t)x;
    *s++ = (j+'0');
    if(i == dez_punkt) {
      if(anzeige_format & 2)
        *s++ = (',');
      else
        *s++ = ('.');
    }  
    x -= j;
    x *= 10;
  }

  if(zeige_periode == WANDEL_D)
    p = u_dim[anzeige_format&1][dimension];
  else
    p = f_dim[anzeige_format&1][dimension];
  while(*p) *s++ = *p++;                    // Dimension anhaengen
  *s++ = 0;                                 // string abschliessen
  *s = 0;                                   // string abschliessen
  s[MAX_ERGEBNIS_STR] = 0;                  // zur Sicherheit begrenzen
}

void Fx_str_mittig(const char *header, double x, char zeige_periode, char *s, uint8_t stellen)    // Anzeige von F, P oder UPM
{
char p[50];
int i,j;
  *p = 0;
  for(i = 0; i < MAX_ERGEBNIS_STR; i++)     // Ergebnisstring loeschen
    s[i] = ' ';
  s[MAX_ERGEBNIS_STR] = 0;                  // Ende von s
  
  wandel_Fx(header,x,zeige_periode,p,stellen);// wandeln
  i = strlen(p);
  i = lcd_breite - i;                       // auf 16 Zeichen/Zeile einmitten
  if(i < 0) i = 0;                          // Unterlauf verhindern
  j = i/2;
  i = 0;
  while(p[i]) s[j++] = p[i++];              // in den Ergebnisstring schreiben
  s[lcd_breite] = 0;                        // Laenge sicher begrenzen
} 

/*
 PIO0-SM0:
 Zeitzaehlung im x-Register mit SystemCoreClock/PIO_TIME_DIV (=4)
 und Ausgabe des Wertes bei negativer Flanke des Eingangssignals
*/
void init_PIO0(void)
{
// PIO0 aktivieren  
  RESETS_CLR->RESET = RESETS_RESET_pio0_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pio0_Msk));

// FIN1_PIN f�r SM0 initialisieren
  IO_BANK0->GPIO2_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // GPIO-Eingang
  PADS_BANK0->GPIO2 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  
  PIO0->SM0_PINCTRL |= 1<<PIO0_SM0_PINCTRL_SIDESET_BASE_Pos | 1<<PIO0_SM0_PINCTRL_SET_BASE_Pos;
  PIO0->SM0_EXECCTRL = (31 << PIO0_SM0_EXECCTRL_WRAP_TOP_Pos) | 
                        (0 << PIO0_SM0_EXECCTRL_WRAP_BOTTOM_Pos) |
                        (FIN1_PIN  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos);  // Pin f�r JMP
// ADR0  
  PIO0->INSTR_MEM0 = 0x0241;      // jmp x-- -> ADR1, 2 x delay
// ADR1
  PIO0->INSTR_MEM1 = 0x00c0;      // jmp FIN1_PIN -> ADR0 warten bis '0'
// input-pin 1 -> 0
  PIO0->INSTR_MEM2 = 0x4020;      // mov x->isr, x-Wert ausgeben
// ADR3
  PIO0->INSTR_MEM3 = 0x0044;      // jmp x-- -> ADR4
// ADR4  
  PIO0->INSTR_MEM4 = 0x00c1;      // jmp FIN1_PIN 0-> 1 dann ADR0
  PIO0->INSTR_MEM5 = 0x0103;      // jmp ADR3, 1 x delay
  
  PIO0_SET->SM0_SHIFTCTRL = PIO0_SM0_SHIFTCTRL_FJOIN_RX_Msk | // FIFO = 8
                            PIO0_SM0_SHIFTCTRL_AUTOPUSH_Msk;  // autopush
  PIO0_SET->CTRL = (1 << (PIO0_CTRL_SM_ENABLE_Pos));          // starten
}


// Zeitstempel erzeugen:
// DMA-Wert = Zeitpunkt, DMA-TRANS-COUNT = Ereignisse
// wenn CH0 fertig wird CH10 angestossen, der wiederum CH0 startet
// daher ist keine ISR notwendig
void init_dma_pio0(void)
{
  RESETS_SET->RESET = RESETS_RESET_dma_Msk;   // reset all CHx
  RESETS_CLR->RESET = RESETS_RESET_dma_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_dma_Msk));

// Zaehlen der Ereignisse und lesen des Zeitstempels  
  DMA->CH0_READ_ADDR = (uint32_t) &PIO0->RXF0;
  DMA->CH0_WRITE_ADDR = (uint32_t) &F1_time_stamp;
  DMA->CH0_TRANS_COUNT = DMA_COUNT;
  DMA->CH0_CTRL_TRIG =  DREQ_PIO0_RX0 << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos |     // PIO0-RX0
                        10 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |  // trigger CH10
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |  // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;               // enable

// Hilfskanal zum Nachtriggern con CH0  
  DMA->CH10_READ_ADDR = (uint32_t) &dma_dummy;                  // dummy location
  DMA->CH10_WRITE_ADDR = (uint32_t) &dma_dummy;
  DMA->CH10_TRANS_COUNT = 1;                                    // single transfer
  DMA->CH10_CTRL_TRIG =  DREQ_AUTO << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos | // free running
                        0 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |   // Trigger CH0  
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |  // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;               // enable
}

// 1 kHz Aufrufe fuer Messzeiten
void PWM_IRQ_WRAP_IRQHandler(void)
{
static uint32_t F1_time_stamp_old;
  if(PWM->INTS & 1) {                           // PWM_CH0?
    PWM->INTR = 1;                              // clear flag
    F1_mess_dauer++;
    F1_timeout_counter++;
    if(F1_time_stamp != F1_time_stamp_old) {    // new count?
      F1_time_stamp_old = F1_time_stamp;        // yes
      F1_timeout_counter = 0;                   // reset timeout
    }
    if(led_count) {                             // ready-LED
      SIO->GPIO_OUT_SET = (1 << LOCAL_LED);
      led_count--;
      if(!led_count)
        SIO->GPIO_OUT_CLR = (1 << LOCAL_LED);
    }  
  }
  __DSB();
}

// ms-Raster fuer Messablauf erzeugen
void init_F1_timer(void)
{
  RESETS_CLR->RESET = RESETS_RESET_pwm_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pwm_Msk));
  
  PWM->CH0_DIV = PWM_VORTEILER << PWM_CH0_DIV_INT_Pos;      //  / 100
  PWM->CH0_TOP = SystemCoreClock/PWM_VORTEILER/MS_TEILER-1; // 1 kHz
  NVIC_EnableIRQ(PWM_IRQ_WRAP_IRQn);
  NVIC_SetPriority(PWM_IRQ_WRAP_IRQn,3);
  irq_tab[PWM_IRQ_WRAP_IRQn] = (uint32_t)PWM_IRQ_WRAP_IRQHandler;  // Vektor im RAM
  PWM->CH0_CSR =  1;                                // enable ms-timer
  PWM->INTE =  1;                                   // enable ISR
}

// Zeitstempel lesen
void get_time_count(uint32_t *time, uint32_t *count)
{
uint32_t temp1, temp2;  
  DMA_CLR->CH0_CTRL_TRIG = DMA_CH0_CTRL_TRIG_EN_Msk;  // stop DMA
  temp1 =   DMA->CH0_AL1_TRANS_COUNT_TRIG;            // count
  temp2 = F1_time_stamp;                              // get time stamp
  DMA_SET->CH0_CTRL_TRIG = DMA_CH0_CTRL_TRIG_EN_Msk;  // continue DMA
  *count = temp1;
  *time  = temp2;
}  
     

void F1_messung(int init_timer)
{
static uint8_t  init = 0, alte_funktion = 255;
uint8_t eff_stellen;
static char p_str[TEMP_STR_LEN];
static char s[TEMP_STR_LEN];
double SM0_clock_temp;

  if(init_timer) { 
    init = 0;                                   // Anzeige + Messung initialisieren
    init_F1_timer();
    init_dma_pio0();
    init_PIO0(); 
  }
   
  if(!init) {
    F1_mess_dauer =  F1_messzeit;             // und 1. Messung sofort abschliessen
    F1_messwert_vorhanden  = 0;               // dabei 1. Ergebnis verwerfen
    F1_timeout_counter = 0; 
    init = 1;
    update_LCD = 1;
  }
  if(!ausgabe_sperre && (anzeige_funktion != alte_funktion || update_LCD)) {  
    update_LCD = 0;
    if(anzeige_funktion > DREHZAHL) 
      anzeige_funktion = F_UND_P;             // sinnvollen Wert einstellen
    alte_funktion = anzeige_funktion;
    loesche_lcd_zeile12();
    switch(anzeige_funktion) {
      case F_UND_P:
        lcd_zeile1((char *)"   Signal F1?   ");
        lcd_zeile2((char *)"   Signal P1?   ");
        break;
      case FREQUENZ:
        lcd_zeile1((char *)"  Frequenz F1   ");
        lcd_zeile2((char *)"    Signal?     ");
        break;
      case PERIODE:
        lcd_zeile1((char *)"   Periode F1   ");
        lcd_zeile2((char *)"    Signal?     ");
        break;
    case DREHZAHL:
        lcd_zeile1((char *)"  Drehzahl F1   ");
        lcd_zeile2((char *)"    Signal?     ");
        break;
    }
  }

  if(F1_timeout_counter >= F1_timeout) {      // bei fehlenden Eingangsimpulsen
    init = 0;                                 // neu initialisieren
  }  

  if(F1_mess_dauer >= F1_messzeit) {          // wenn Mindestzeit erreicht
    get_time_count(&F1_end_zeit, &F1_end_ereignis);
    F1_mess_zeit = F1_start_zeit-F1_end_zeit;
    F1_start_zeit = F1_end_zeit;
    F1_mess_ereignisse = (F1_start_ereignis-F1_end_ereignis) & DMA_COUNT_MASKE;
    F1_start_ereignis = F1_end_ereignis;
    if(F1_mess_ereignisse) {
      SM0_clock_temp = (SystemCoreClock+f_ref_korrektur)/PIO_TIME_DIV;
      F1_ergebnis = (double)F1_mess_ereignisse/(double)F1_mess_zeit * SM0_clock_temp;
      F1_mess_dauer = 0;                      // neue Messung beginnen
      if(F1_messwert_vorhanden) { 
        eff_stellen = F1_stellen;
        if(F1_vorteiler_aktiv)
          F1_ergebnis *= (double)F1_faktor;   // und ggf. skalieren

        led_count = led_zeit;
// serielle Datenausgabe zuerst        
        if(ser_ausgabe_kanal == P1) {
          wandel_Fx("",F1_ergebnis, WANDEL_P, s, eff_stellen);
          sende_USB(s);
        }
        if(ser_ausgabe_kanal == UPM1) {
          wandel_Fx("",F1_ergebnis*60.0/F1_upm_divisor, WANDEL_D, s, eff_stellen);
          sende_USB(s);
        }
        wandel_Fx("",F1_ergebnis, WANDEL_F, s, eff_stellen);
        if(ser_ausgabe_kanal == F1) 
          sende_USB(s);
// danach Anzeige auf LCD        
        switch(anzeige_funktion) {
          case F_UND_P:
            Fx_str_mittig("F = ",F1_ergebnis, WANDEL_F, p_str, eff_stellen);
            messwert_lcd1(p_str);
            Fx_str_mittig("P = ",F1_ergebnis, WANDEL_P, p_str, eff_stellen);
            messwert_lcd2(p_str);
            break;
          case FREQUENZ:
            Fx_str_mittig("",F1_ergebnis, WANDEL_F, p_str, eff_stellen);
            messwert_lcd2(p_str);
            break;
          case PERIODE:
            Fx_str_mittig("",F1_ergebnis, WANDEL_P, p_str, eff_stellen);
            messwert_lcd2(p_str);
            break;
          case DREHZAHL:
            Fx_str_mittig("",F1_ergebnis*60.0/F1_upm_divisor, WANDEL_D, s, eff_stellen);
            messwert_lcd2(s);
            break;
        }
      } else F1_messwert_vorhanden  = 1;  // sperre wieder aufheben
    }
  }  
}  
